<'
# Copyright (c) 2012-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Sat Jan 14 2012
#

if {[llength $argv] != 2} { error "expected arguments: component task" }
lassign $argv component task

# compute handy shortcuts
set comp [$component name]
set tname [$task name]

lang c
'>
#include "autoconf/acheader.h"

#include <sys/time.h>

#include "<"$comp">_internals.h"


/* --- local data ---------------------------------------------------------- */

static int	<"$comp">_<"$tname">_invoke(genom_event &state,
                         genom_component_data *self);


/* <"[--- ${comp}_${tname}_task -----------------------------------------]"> */

/** <"$tname"> task of component <"$comp">
 *
 */

void *
<"$comp">_<"$tname">_task(void *data)
{
  genom_component_data *self = (genom_component_data *)data;
  genom_exec_task &task = self->tasks.<"$tname">;
  genom_state_component *state = self->ports.genom_state.handle.data(&task.context);
  struct timeval tvs, tve;
  genom_activity *a;
  int sched, shutdown, done;
  size_t i;
  int s;

  pthread_mutex_lock(&self->task_lock);
  task.status = genom_ok;
  task.pact = <"$comp">_start;

  /* permanent activity start codel */
  s = <"$comp">_<"$tname">_invoke(task.pact, self);
  if (s) task.status = task.pact;

  genom_log_info("<"$tname"> task initialized and running");
  pthread_cond_signal(&task.signal);
  pthread_mutex_unlock(&self->task_lock);

  if (task.status) {
    genom_log_warn("shutting down <"$tname"> task");
    return NULL;
  }

  /* loop */
  done = shutdown = 0;
  pthread_mutex_lock(&self->task_lock);
  do {
    while(!task.wakeup && !genom_shutdown) {
      s = pthread_cond_wait(&task.signal, &self->task_lock);
      if (s) err(2, "timer task <"$tname">");
    }
    task.runnable = task.wakeup = false;
    sched = 0;
    genom_log_debug("<"$tname"> tick %d", self->ticks);
    gettimeofday(&tvs, NULL);

    /* handle shutdown: interrupt all if not done already */
    if (genom_shutdown && !shutdown) {
      shutdown = 1;
      task.pact = <"$comp">_stop;
      for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
        a = self->activities.activities[i];
        if (!a || a->task != task.id) continue;
        a->stop("kill");
      }
    }

    /* handle permanent activity */
    s = <"$comp">_<"$tname">_invoke(task.pact, self);
    if (s) {
      task.status = task.pact; break;
    }

    /* handle current activities for this task */
    for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
      a = self->activities.activities[i];
      if (!a || a->task != task.id) continue;
      if (a->status == ACT_STOP) {
        a->status = ACT_STOPPING;
        a->codel = <"${comp}">_stop;
      }
      if (a->status != ACT_RUNNING && a->status != ACT_STOPPING)
        continue;

      genom_log_debug("<"$tname"> activity %d", a->id);
      a->codel = a->invoke();
      if (a->codel == NULL || a->codel == <"$comp">_ether) {
        if (a->status == ACT_RUNNING)
          a->status = ACT_ETHER;
        else
          a->status = ACT_STOPPED;
        if (a->codel == NULL)
          a->exdetail = (void *)task.context.raised(&a->codel, &task.context);
        <"$comp">_cntrl_task_signal(a);
      } else if (a->codel != <"$comp">_sleep || a->status == ACT_STOP)
        sched = 1;

      /* update state port */
      if (a->stateid >= 0) {
        strncpy(state->activity._buffer[a->stateid].state,
                a->codel ? a->codel : "",
                sizeof(state->activity._buffer[a->stateid].state));
      }
    }
    if (sched) self->exec_task_signal();

    /* update state port */
    gettimeofday(&tve, NULL);

    state->task._buffer[task.stateid].cputime.last =
      tve.tv_sec - tvs.tv_sec + (tve.tv_usec - tvs.tv_usec)*1e-6;

    if (state->task._buffer[task.stateid].cputime.last >
        state->task._buffer[task.stateid].cputime.max)
      state->task._buffer[task.stateid].cputime.max =
        state->task._buffer[task.stateid].cputime.last;

    state->task._buffer[task.stateid].cputime.avg *= 0.9;
    state->task._buffer[task.stateid].cputime.avg +=
      0.1 * state->task._buffer[task.stateid].cputime.last;
    self->ports.genom_state.publish();

    pthread_mutex_unlock(&self->task_lock);
    sched_yield();
    pthread_mutex_lock(&self->task_lock);

    /* handle shutdown: continue until there is some activtiy, or break */
    if (genom_shutdown) {
      done = 1;
      if (task.pact != <"$comp">_ether) done = 0;
      for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
        a = self->activities.activities[i];
        if (!a || a->task != task.id) continue;
        if (a->status != ACT_ETHER && a->status != ACT_STOPPED) done = 0;
      }
    }
  } while(!done);

  pthread_mutex_unlock(&self->task_lock);

  genom_log_info("shutting down <"$tname"> task");
  return NULL;
}


/* === task codels ========================================================= */
<'
foreach e [$task fsm] {
  set codel [$task fsm $e]
'>

/* state <"$e">  */
static genom_event
<"$comp">_tcodel_<"$tname">_<"[$e cname]">(genom_component_data *self)
{
  genom_event s;
<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom::ids::pinit(<"[$p name]_">);
  genom::ids::pcopy(<"[$p name]_">, self->ids<"[$p member]">);
<'    }'>
<'  }'>

<'  # build parameter list
  set plist [list]
  foreach p [$codel parameters] {
    switch -glob -- "[$codel kind]/[$p src]" {
      */local	{ error "task codel with service parameter" }
      */remote	{ set m "self->remotes.[[$p remote] name].handle" }
      */port	{ set m "self->ports.[[$p port] name].handle" }
      async/ids	{ set m [$p name]_ }
      sync/ids	{ set m "self->ids[$p member]" }
      default	{ error "reached unreachable code" }
    }

    switch  -glob -- "[$p dir]/[$p src]" {
      in/* - */port {
        lappend plist [[$p type] pass value $m]
      }
      default {
        lappend plist [[$p type] pass reference $m]
      }
    }
  }
  lappend plist "&self->tasks.${tname}.context"
'>
<'if {[$codel kind] == "async"} {'>
  pthread_mutex_unlock(&self->task_lock);
<'}'>
  s = <"[$codel invoke $plist]">;
<'if {[$codel kind] == "async"} {'>
  pthread_mutex_lock(&self->task_lock);

<'  foreach p [$codel parameters ids] {'>
<'    if {[$p dir] != "in"} {'>
  genom::ids::pcopy(self->ids<"[$p member]">, <"[$p name]_">);
<'    }'>
  genom::ids::pfini(<"[$p name]_">);
<'  }'>
<'}'>
  return s;
}
<'}'>


/* invoke task codels according to state */
static int
<"$comp">_<"$tname">_invoke(genom_event &state, genom_component_data *self)
{
  if (state == <"$comp">_ether) return 0;

<'  foreach e [$task fsm] {'>
  if (state == <"[$e cname]">) {
    genom_event s;
    s = <"$comp">_tcodel_<"$tname">_<"[$e cname]">(self);
    if (
<'    foreach y [[$task fsm $e] yield] {'>
      s == <"[$y cname]"> ||
<'    }'>
      0) {
      state = s;
<'    foreach p [[$task fsm $e] parameters port] {'>
<'      if {"handle" ni [[$p port] kind] && "out" in [[$p port] kind]} {'>
      self->ports.<"[$p name]">.publish();
<'      }'>
<'    }'>
      if (s != <"$comp">_sleep) self->exec_task_signal();
      return 0;
    }
<'    if {[llength [$task throw]]} {'>
    if (
<'      foreach t [$task throw] {'>
      s == <"[$t cname]">_id ||
<'      }'>
      0) {
      state = s;
      return -1;
    }
<'    }'>

    genom_bad_transition_detail d;
    strncpy(d.from, state, sizeof(d.from));
    d.from[sizeof(d.from)-1] = *"";
    strncpy(d.to, s, sizeof(d.to));
    d.to[sizeof(d.to)-1] = *"";
    genom_log_warn(
      "bad transition from %s to %s in task %s", state, s, "<"$tname">");
    state = genom_bad_transition(&d, &self->tasks.<"$tname">.context);
    return -1;
  }
<'  }'>
  /* default actions */
  if (state == <"$comp">_start || state == <"$comp">_stop) {
    state = <"$comp">_ether;
    return 0;
  }
  if (state == <"$comp">_sleep) return 0;

  genom_log_warn("bad state %s in task %s permanent activity",
                 state, "<"$tname">");
  assert(!"internal error");
  return -1;
}


/* === services codels ===================================================== */

<'foreach s [$task services] {'>
/* <"[--- Service [$s name] ---------------------------------------------]"> */

<'  foreach e [$s fsm] { set codel [$s fsm $e]'>

/* state <"[$e name]"> */
static inline genom_event
<"$comp">_codel_<"[$s name]">_<"[$e cname]">(
  genom_activity_data_<"[$s name]"> *a, genom_component_data *self)
{
  genom_event s;

<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom::ids::pinit(<"[$p name]_">);
  genom::ids::pcopy(<"[$p name]_">, self->ids<"[$p member]">);
<'    }'>
<'  }'>

<'    # build parameter list
    set plist [list]
    foreach p [$codel parameters] {
      switch -glob -- "[$codel kind]/[$p src]/[$p dir]" {
        */local/*	{
          switch -- [[$p param] dir] {
            local	{ set m "a->locals.[$p name]" }
            in		{ set m "a->in.[$p name]" }
            inout - out { set m "a->out.[$p name]" }
            default	{ error "unsupported local direction" }
          }
        }
        */remote/*	{ set m "self->remotes.[[$p remote] name].handle" }
        */port/*	{ set m "self->ports.[[$p port] name].handle" }
        async/ids/*	{ set m [$p name]_ }
        sync/ids/*	{ set m "self->ids[$p member]" }
        default		{ error "reached unreachable code" }
      }

      switch  -glob -- "[$p dir]/[$p src]" {
	in/* - */port {
          lappend plist [[$p type] pass value $m]
        }
	default {
	  lappend plist [[$p type] pass reference $m]
	}
      }
    }
    lappend plist "&self->tasks.${tname}.context"
'>
<'    if {[$codel kind] == "async"} {'>
  pthread_mutex_unlock(&self->task_lock);
<'    }'>
  s = <"[$codel invoke $plist]">;
<'    if {[$codel kind] == "async"} {'>
  pthread_mutex_lock(&self->task_lock);

<'      foreach p [$codel parameters ids] {'>
<'        if {[$p dir] != "in"} {'>
  genom::ids::pcopy(self->ids<"[$p member]">, <"[$p name]_">);
<'        }'>
  genom::ids::pfini(<"[$p name]_">);
<'      }'>
<'    }'>
  return s;
}

<'  }'>
/* invoke codels according to <"[$s name]"> state */
template<> genom_event
genom_activity_data_<"[$s name]">::invoke()
{
  genom_event s;

  genom_log_debug(
    "task %s invoking activity %s", "<"$tname">", "<"[$s name]">");

<'  foreach e [$s fsm] {'>
  if (codel == <"[$e cname]">) {
    s = <"$comp">_codel_<"[$s name]">_<"[$e cname]">(this, self->data->self);
    genom_log_debug("service %s in task %s yielded %s",
                    "<"[$s name]">", "<"$tname">", s);
    if (
<'    foreach y [[$s fsm $e] yield] {'>
        s == <"[$y cname]"> ||
<'    }'>
        0) {
      return s;
    }
    goto ex;
  }
<'  }'>
  /* default events, if not already handled above */
  if (codel == <"$comp">_sleep) return <"$comp">_sleep;
  if (codel == <"$comp">_stop) return <"$comp">_ether;

  /* this cannot happen by construction */
  genom_log_warn("bad state %s in service %s", codel, "<"[$s name]">");
  assert(!"bad state");
  abort();
  return NULL;

ex:
  if (
<'  foreach t [$s throw] {'>
    s != <"[$t cname]">_id &&
<'  }'>
    1) {
    genom_bad_transition_detail d;
    strncpy(d.from, codel, sizeof(d.from)); d.from[sizeof(d.from)-1] = *"";
    strncpy(d.to, s, sizeof(d.to)); d.to[sizeof(d.to)-1] = *"";
    genom_log_warn("bad transition from %s to %s in service %s",
                   codel, s?s:"genom_ok", "<"[$s name]">");
    genom_bad_transition(&d, self);
  }

  return NULL;
}

<'}'>
